# Copyright (c) 2012 David Christian

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial
# portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
# OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# -------------------------------------------------------------------------------------------------------------


# This is the main game file. Run this to play the game :)

# This little game was created for the PyWeek competition of May 2012. It was put together in my spare time during a week,
# so please forgive the rough and messy nature of the code in places :)

import pygame
from pygame.locals import *
import time, os
from conwaygrid import ConwayGrid
from conwayui import ConwayUI

GRID_W = 100
GRID_H = 100
GRID_PIX_SIZE = 6
SCREEN_W = GRID_W * GRID_PIX_SIZE
SCREEN_H = (GRID_H * GRID_PIX_SIZE) + (GRID_PIX_SIZE * 8)

SHAPE_A = 0
SHAPE_B = 1
SHAPE_C = 2
SHAPE_D = 3
SHAPE_E = 4
SHAPE_F = 5

FPS = 60

GAME_TITLE_MODE = 0
GAME_SETUP_MODE = 1
GAME_SIMULATION_MODE = 2
GAME_FREEPLAY_MODE = 3
GAME_OVER_MODE = 4

# These shapes are all 5x5. 0=off 1=on
grid_shape_a = [[0,0,0,0,0],
           [0,0,0,0,0],
           [0,0,1,0,0],
           [0,0,0,0,0],
           [0,0,0,0,0]]

grid_shape_b = [[0,0,0,0,0],
           [0,1,0,0,0],
           [0,1,0,1,0],
           [0,1,1,0,0],
           [0,0,0,0,0]]

grid_shape_c = [[0,0,0,0,0],
           [0,0,1,0,0],
           [0,0,1,0,0],
           [0,0,1,0,0],
           [0,0,0,0,0]]

grid_shape_d = [[0,0,0,0,0],
           [0,1,1,1,0],
           [0,1,1,1,0],
           [0,1,1,1,0],
           [0,0,0,0,0]]

grid_shape_e = [[0,0,0,0,0],
           [0,0,0,1,0],
           [0,1,0,1,0],
           [0,0,1,0,1],
           [0,0,1,0,0]]

grid_shape_f = [[0,0,0,0,0],
           [0,0,0,1,0],
           [0,1,0,0,1],
           [0,1,0,0,1],
           [0,0,1,0,0]]

play_shape = [[0,0,0,0,0],
           [1,1,1,1,1],
           [1,0,0,0,1],
           [0,1,0,1,0],
           [0,0,1,0,0]]

pause_shape = [[0,0,0,0,0],
           [0,1,1,1,0],
           [0,0,0,0,0],
           [0,1,1,1,0],
           [0,0,0,0,0]]

stop_shape = [[1,0,0,0,1],
           [0,1,0,1,0],
           [0,0,1,0,0],
           [0,1,0,1,0],
           [1,0,0,0,1]]

# There are now FOUR! multidemensional arrays. There are probably much
# more efficient ways to do this but with the deadline just over 3 days away
# at time of writing I am sticking to this method of storing cell data. 
grid = []
grid_temp = []
virus_grid = []
virus_temp = []

# The 'grid' arrays hold the cell data and the 'temp' arrays hold
# the number of each cells neighbouring active cells
player_grid = [0] * GRID_H
for i in range(GRID_H):
    player_grid[i] = [0] * GRID_W

player_temp = [0] * GRID_H
for i in range(GRID_H):
    player_temp[i] = [0] * GRID_W

virus_grid = [0] * GRID_H
for i in range(GRID_H):
    virus_grid[i] = [0] * GRID_W

virus_temp = [0] * GRID_H
for i in range(GRID_H):
    virus_temp[i] = [0] * GRID_W

pygame.init()
screen = pygame.display.set_mode([SCREEN_W, SCREEN_H])
pygame.display.set_caption("MadScienceOS")
clock = pygame.time.Clock()

conway_grid = ConwayGrid(GRID_W, GRID_H, GRID_PIX_SIZE)

# Prepare various pygame surfaces for blitting later
conway_img = pygame.Surface((SCREEN_W, SCREEN_H))
conway_img.fill ((0,0,0)) 
active_img = pygame.Surface((GRID_PIX_SIZE, GRID_PIX_SIZE))
active_img.fill((0,255,0))
virus_img = pygame.Surface((GRID_PIX_SIZE, GRID_PIX_SIZE))
virus_img.fill((0,100,255))
mouse_img = pygame.Surface((GRID_PIX_SIZE, GRID_PIX_SIZE))
mouse_img.fill((255,255,0))
ui_img = pygame.Surface((GRID_PIX_SIZE, GRID_PIX_SIZE))
ui_img.fill((128,128,128))
ui_rollover_img = pygame.Surface((GRID_PIX_SIZE, GRID_PIX_SIZE))
ui_rollover_img.fill((255,255,0))
ui_selected_img = pygame.Surface((GRID_PIX_SIZE, GRID_PIX_SIZE))
ui_selected_img.fill((255,255,255))

title_font = pygame.font.Font(os.path.join("data", "bitwise.ttf"), 86)
game_font = pygame.font.Font(os.path.join("data", "bitwise.ttf"), 16)

button_experiment_img = pygame.Surface((300, 50),pygame.SRCALPHA)
button_experiment_rollover = pygame.Surface((300, 50),pygame.SRCALPHA)
button_experiment_img.fill((0,0,175, 150))
button_experiment_img.blit(game_font.render("Start Experimental Mode", 0, ((255, 255, 255))), (53, 18))
button_experiment_rollover.fill((0,0,175, 200))
button_experiment_rollover.blit(game_font.render("Start Experimental Mode", 0, ((255, 255, 255))), (53, 18))

button_battle_img = pygame.Surface((300, 50),pygame.SRCALPHA)
button_battle_rollover = pygame.Surface((300, 50),pygame.SRCALPHA)
button_battle_img.fill((0,0,175, 150))
button_battle_img.blit(game_font.render("Start Battle Mode", 0, ((255, 255, 255))), (81, 18))
button_battle_rollover.fill((0,0,175, 200))
button_battle_rollover.blit(game_font.render("Start Battle Mode", 0, ((255, 255, 255))), (81, 18))

button_shutdown_img = pygame.Surface((300, 50),pygame.SRCALPHA)
button_shutdown_rollover = pygame.Surface((300, 50),pygame.SRCALPHA)
button_shutdown_img.fill((0,0,175, 150))
button_shutdown_img.blit(game_font.render("Exit MadScienceOS", 0, ((255, 255, 255))), (77, 18))
button_shutdown_rollover.fill((0,0,175, 200))
button_shutdown_rollover.blit(game_font.render("Exit MadScienceOS", 0, ((255, 255, 255))), (77, 18))

button_loser_img = pygame.Surface((300, 50),pygame.SRCALPHA)
button_loser_rollover = pygame.Surface((300, 50),pygame.SRCALPHA)
button_loser_img.fill((0,0,175, 150))
button_loser_img.blit(game_font.render("Your cells were defeated", 0, ((255, 255, 255))), (57, 18))
button_loser_rollover.fill((0,0,175, 200))
button_loser_rollover.blit(game_font.render("Your cells were defeated", 0, ((255, 255, 255))), (57, 18))

button_winner_img = pygame.Surface((300, 50),pygame.SRCALPHA)
button_winner_rollover = pygame.Surface((300, 50),pygame.SRCALPHA)
button_winner_img.fill((0,175,0, 150))
button_winner_img.blit(game_font.render("Your cells were victorious!", 0, ((255, 255, 255))), (57, 18))
button_winner_rollover.fill((0,175,0, 200))
button_winner_rollover.blit(game_font.render("Your cells were victorious!", 0, ((255, 255, 255))), (57, 18))

# Setup the buttons
conway_ui = ConwayUI(ui_img, ui_rollover_img, ui_selected_img, GRID_PIX_SIZE)
button_a = conway_ui.create_grid_button(0, (GRID_H - 1) * GRID_PIX_SIZE, grid_shape_a, 1)
button_b = conway_ui.create_grid_button((9 * GRID_PIX_SIZE), (GRID_H - 1) * GRID_PIX_SIZE, grid_shape_b, 1)
button_c = conway_ui.create_grid_button((9 * GRID_PIX_SIZE) + (9 * GRID_PIX_SIZE), (GRID_H - 1) * GRID_PIX_SIZE, grid_shape_c, 1)
button_d = conway_ui.create_grid_button((18 * GRID_PIX_SIZE) + (9 * GRID_PIX_SIZE), (GRID_H - 1) * GRID_PIX_SIZE, grid_shape_d, 1)
button_e = conway_ui.create_grid_button((27 * GRID_PIX_SIZE) + (9 * GRID_PIX_SIZE), (GRID_H - 1) * GRID_PIX_SIZE, grid_shape_e, 1)
button_f = conway_ui.create_grid_button((36 * GRID_PIX_SIZE) + (9 * GRID_PIX_SIZE), (GRID_H - 1) * GRID_PIX_SIZE, grid_shape_f, 1)

button_title_experiment = conway_ui.create_image_button(25 * GRID_PIX_SIZE, 50 * GRID_PIX_SIZE, button_experiment_img, button_experiment_rollover, 2)
button_title_battle = conway_ui.create_image_button(25 * GRID_PIX_SIZE, 40 * GRID_PIX_SIZE, button_battle_img, button_battle_rollover, 2)
button_title_shutdown = conway_ui.create_image_button(25 * GRID_PIX_SIZE, 65 * GRID_PIX_SIZE, button_shutdown_img, button_shutdown_rollover, 2)

button_loser = conway_ui.create_image_button(25 * GRID_PIX_SIZE, 50 * GRID_PIX_SIZE, button_loser_img, button_loser_rollover, 3)
button_winner = conway_ui.create_image_button(25 * GRID_PIX_SIZE, 50 * GRID_PIX_SIZE, button_winner_img, button_winner_rollover, 4)

play_button = conway_ui.create_grid_button(((GRID_W - 36) * GRID_PIX_SIZE) + (9 * GRID_PIX_SIZE), (GRID_H - 1) * GRID_PIX_SIZE, play_shape ,1)
pause_button = conway_ui.create_grid_button(((GRID_W - 27) * GRID_PIX_SIZE) + (9 * GRID_PIX_SIZE), (GRID_H - 1) * GRID_PIX_SIZE, pause_shape, 1)
stop_button = conway_ui.create_grid_button(((GRID_W - 18) * GRID_PIX_SIZE) + (9 * GRID_PIX_SIZE), (GRID_H - 1) * GRID_PIX_SIZE, stop_shape, 1)

generation = 0
mouse_x = 0
mouse_y = 0
main_loop_running = True
game_mode = GAME_TITLE_MODE
current_shape = SHAPE_B
selected_shape = grid_shape_b
single_cell_mode = True
last_cell_collision = 0
allocated_cells = 150
game_paused = False

conway_ui.set_group_visibility(2)

# Setup the virus grid for the title screen
conway_grid.populate_virus_grid(virus_grid, 35, grid_shape_b)
conway_grid.populate_virus_grid(virus_grid, 25, grid_shape_c)
conway_grid.populate_virus_grid(virus_grid, 7, grid_shape_d)
conway_grid.populate_virus_grid(virus_grid, 25, grid_shape_e)
conway_grid.populate_virus_grid(virus_grid, 25, grid_shape_f)

# Show a cute little intro! 
screen.fill((0,0,0))
for i in range(5):
    dots = "." * i
    screen.fill((0,0,0))
    screen.blit(game_font.render("Booting MadScienceOS" + dots, 0, ((255, 255, 255))), (215, 300))
    pygame.display.flip()  
    time.sleep(0.5)    

# MAIN GAME PROCESSING LOOP
# -------------------------
while main_loop_running == True:
    conway_img.fill((0,0,0))

    # Do some title screen stuff
    if game_mode == GAME_TITLE_MODE:
        conway_grid.populate_virus_grid(virus_grid, 5, grid_shape_b)
        conway_grid.populate_virus_grid(virus_grid, 6, grid_shape_e)
        conway_grid.populate_virus_grid(virus_grid, 9, grid_shape_d)
        
    mouse_x, mouse_y = pygame.mouse.get_pos()
    mx = mouse_x / GRID_PIX_SIZE * GRID_PIX_SIZE
    my = mouse_y / GRID_PIX_SIZE * GRID_PIX_SIZE

    # Process various pygame events such as mouse presses etc.
    for e in pygame.event.get():            
        if e.type == KEYDOWN:
            if e.key == pygame.K_ESCAPE:
                main_loop_running = False

            if e.key == pygame.K_x:
            # Just for now, place a 'virus' blinker at the mouse position
                virus_grid[(mouse_x / GRID_PIX_SIZE)][(mouse_y / GRID_PIX_SIZE)] = 1               
                virus_grid[(mouse_x / GRID_PIX_SIZE)][(mouse_y / GRID_PIX_SIZE)+1] = 1
                virus_grid[(mouse_x / GRID_PIX_SIZE)][(mouse_y / GRID_PIX_SIZE)+2] = 1

        if e.type == pygame.QUIT:
            main_loop_running = False
            
        if e.type == MOUSEBUTTONDOWN and (game_mode == GAME_SETUP_MODE or game_mode == GAME_FREEPLAY_MODE):            
            if mouse_x > 0 and mouse_y > 0 and mouse_x < SCREEN_W and mouse_y < (GRID_H * GRID_PIX_SIZE):                              
                # Place current shape on the player grid
                if single_cell_mode == True:
                    if pygame.mouse.get_pressed()[0] == True:
                        if allocated_cells > 0:
                            if  player_grid[mx / GRID_PIX_SIZE][my / GRID_PIX_SIZE] == 0:
                                allocated_cells -= 1
                            player_grid[mx / GRID_PIX_SIZE][my / GRID_PIX_SIZE] = 1
                    else:
                        if player_grid[mx / GRID_PIX_SIZE][my / GRID_PIX_SIZE] == 1:
                            allocated_cells += 1
                        player_grid[mx / GRID_PIX_SIZE][my / GRID_PIX_SIZE] = 0
                else:                    
                    for y in range(0, 5):
                        for x in range(0, 5):
                            if selected_shape[x][y] > 0 and virus_grid[mx / GRID_PIX_SIZE][my / GRID_PIX_SIZE] == 0:                                
                                if (mx / GRID_PIX_SIZE) + x < GRID_W and (my / GRID_PIX_SIZE) + y < GRID_H:                        
                                    if pygame.mouse.get_pressed()[0] == True:
                                        if allocated_cells > 0:
                                            if player_grid[(mx / GRID_PIX_SIZE) + x][(my / GRID_PIX_SIZE) + y] == 0:
                                                allocated_cells -= 1
                                            player_grid[(mx / GRID_PIX_SIZE) + x][(my / GRID_PIX_SIZE) + y] = 1                                            
                                    else:
                                        if player_grid[(mx / GRID_PIX_SIZE) + x][(my / GRID_PIX_SIZE) + y] > 0:
                                            allocated_cells += 1
                                        player_grid[(mx / GRID_PIX_SIZE) + x][(my / GRID_PIX_SIZE) + y] = 0                                            
                                            

    # Check for ui button presses
    button_pressed = conway_ui.check_buttons(mouse_x, mouse_y, pygame.mouse.get_pressed())
    
    if game_mode == GAME_SETUP_MODE or game_mode == GAME_FREEPLAY_MODE:
        if button_pressed == button_a:
            single_cell_mode = True        
        elif button_pressed == button_b:
            selected_shape = grid_shape_b
            single_cell_mode = False
        elif button_pressed == button_c:
            selected_shape = grid_shape_c
            single_cell_mode = False
        elif button_pressed == button_d:
            selected_shape = grid_shape_d
            single_cell_mode = False
        elif button_pressed == button_e:
            selected_shape = grid_shape_e
            single_cell_mode = False
        elif button_pressed == button_f:
            selected_shape = grid_shape_f
            single_cell_mode = False

    if game_mode == GAME_SETUP_MODE or game_mode == GAME_FREEPLAY_MODE or game_mode == GAME_SIMULATION_MODE: 
        if button_pressed == pause_button:
            game_paused = True
        elif button_pressed == play_button:
            if game_paused:
                game_paused = False
            elif game_mode == GAME_SETUP_MODE:                
                generation = 0
                last_cell_collision = 0
                game_mode = GAME_SIMULATION_MODE
        elif button_pressed == stop_button:
            game_mode = GAME_TITLE_MODE
            conway_grid.clear_grid(virus_grid)
            conway_grid.clear_grid(player_grid)
            conway_grid.populate_virus_grid(virus_grid, 35, grid_shape_b)
            conway_grid.populate_virus_grid(virus_grid, 25, grid_shape_c)
            conway_grid.populate_virus_grid(virus_grid, 7, grid_shape_d)
            conway_grid.populate_virus_grid(virus_grid, 25, grid_shape_e)
            conway_grid.populate_virus_grid(virus_grid, 25, grid_shape_f)
            game_paused = False
            conway_ui.set_group_visibility(2)

    if game_mode == GAME_OVER_MODE:        
        if button_pressed == button_loser or button_pressed == button_winner:            
            game_mode = GAME_TITLE_MODE
            conway_ui.set_group_visibility(2)            
            conway_grid.clear_grid(player_grid)
            conway_grid.clear_grid(virus_grid)
            conway_grid.populate_virus_grid(virus_grid, 35, grid_shape_b)
            conway_grid.populate_virus_grid(virus_grid, 25, grid_shape_c)
            conway_grid.populate_virus_grid(virus_grid, 7, grid_shape_d)
            conway_grid.populate_virus_grid(virus_grid, 25, grid_shape_e)
            conway_grid.populate_virus_grid(virus_grid, 25, grid_shape_f)            
            generation = 0
            last_cell_collision = 0
            paused = False            
            time.sleep(0.5)
            
    if game_mode == GAME_TITLE_MODE:
        if button_pressed == button_title_experiment:
            generation = 0
            game_mode = GAME_FREEPLAY_MODE            
            single_cell_mode = True
            last_cell_collision = 0
            allocated_cells = 100000
            game_paused = True
            conway_ui.set_group_visibility(1)
            conway_grid.clear_grid(player_grid)
            conway_grid.clear_grid(virus_grid)
            conway_ui.set_selected_button(button_a)
        elif button_pressed == button_title_battle:            
            generation = 0
            game_mode = GAME_SETUP_MODE            
            single_cell_mode = True
            conway_ui.set_selected_button(button_a)
            last_cell_collision = 0
            allocated_cells = 200            
            conway_ui.set_group_visibility(1)            
            conway_grid.clear_grid(player_grid)
            conway_grid.clear_grid(virus_grid)
            conway_grid.populate_virus_grid(virus_grid, 50, grid_shape_b)
            conway_grid.populate_virus_grid(virus_grid, 50, grid_shape_c)
            conway_grid.populate_virus_grid(virus_grid, 50, grid_shape_d)
            conway_grid.populate_virus_grid(virus_grid, 50, grid_shape_e)
            conway_grid.populate_virus_grid(virus_grid, 50, grid_shape_f)
            conway_ui.set_selected_button(button_a)
        elif button_pressed == button_title_shutdown:
            main_loop_running = False                   
               
    # Which mode are we running in?
    if not game_paused:
        if game_mode == GAME_SIMULATION_MODE or game_mode == GAME_FREEPLAY_MODE:
            conway_grid.process_life_grid(player_grid, player_temp)

    # Game processing specific to simulation (battle) mode
    if (game_mode == GAME_SIMULATION_MODE or game_mode == GAME_TITLE_MODE) and not game_paused:
        conway_grid.process_life_grid(virus_grid, virus_temp)    

    # Check for battle game over condition
    if game_mode == GAME_SIMULATION_MODE:
        # Has the simulation run its course?
        if conway_grid.process_grid_collisions(player_grid, player_temp, virus_grid, virus_temp):
            last_cell_collision = generation    

        # It has if we have gone 200 generations with no cell collisions        
        if (last_cell_collision > 0 and last_cell_collision < generation - 100) or (last_cell_collision == 0 and generation > 500):            
            green_cells = 0
            virus_cells = 0
            for y in range(0, GRID_H - 1):
                for x in range(0, GRID_W - 1):
                    if player_grid[x][y] > 0:
                        green_cells += 1
                    if virus_grid[x][y] > 0:
                        virus_cells += 1

            game_mode = GAME_OVER_MODE  
            # Who won?!
            if green_cells >= virus_cells:
                conway_ui.set_group_visibility(4)
            else:
                conway_ui.set_group_visibility(3)
                        
    # Draw the grid
    for y in range(1, GRID_H - 1):
        for x in range(1, GRID_W - 1):    
            if player_grid[x][y] > 0:
                conway_img.blit(active_img, ((x * GRID_PIX_SIZE, y * GRID_PIX_SIZE)))            
            if virus_grid[x][y] > 0:
                conway_img.blit(virus_img, ((x * GRID_PIX_SIZE, y * GRID_PIX_SIZE)))

    # Draw mouse cell placement cursor
    if game_mode == GAME_SETUP_MODE or game_mode == GAME_FREEPLAY_MODE:        
        if mouse_x > 0 and mouse_x < GRID_W * GRID_PIX_SIZE and mouse_y > 0 and mouse_y < GRID_H * GRID_PIX_SIZE:            
            if single_cell_mode == True:
                conway_img.blit(mouse_img, (mx, my))
            else:
                # Draw the shape
                if mouse_y < GRID_H * GRID_PIX_SIZE:
                    for y in range(0, 5):
                        for x in range(0, 5):
                            if selected_shape[x][y] > 0:
                                conway_img.blit(mouse_img, (mx + GRID_PIX_SIZE * x, my + GRID_PIX_SIZE * y))

    if game_mode <> GAME_TITLE_MODE and not game_mode == GAME_OVER_MODE:
        conway_ui.draw_all(conway_img, mouse_x, mouse_y)
    
    # Draw some gridlines
    for y in range(0, GRID_H + 8):
        pygame.draw.line(conway_img, (0,0,50,128), (0, y * GRID_PIX_SIZE), (SCREEN_W, y * GRID_PIX_SIZE))        
    for x in range(0, GRID_W):
        pygame.draw.line(conway_img, (0,0,50,128), (x * GRID_PIX_SIZE, 0), (x * GRID_PIX_SIZE, SCREEN_H))    

    # Various game mode stuff
    if game_mode == GAME_OVER_MODE:        
        conway_ui.draw_all(conway_img, mouse_x, mouse_y)    

    if game_mode == GAME_TITLE_MODE:
        conway_img.blit(title_font.render("CONWAY", 0, ((255, 255, 255))), (120, 100))    
        conway_ui.draw_all(conway_img, mouse_x, mouse_y)
        
    if game_mode == GAME_SETUP_MODE:
        conway_img.blit(game_font.render("Remaining Cells: " + str(allocated_cells), 0, ((255, 255, 255))), (10, 10))
        
    # Display the frame
    screen.blit(conway_img,(0,0))
    pygame.display.flip()    
    clock.tick(FPS)

    generation += 1          
    
pygame.quit()
