# Copyright (c) 2012 David Christian

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial
# portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
# OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import pygame
from pygame.locals import *

BUTTON_ID = 0
BUTTON_GROUP = 1
BUTTON_X = 2
BUTTON_Y = 3
BUTTON_SHAPE = 4
BUTTON_IMAGE = 4 # 4th array element will be image for image button and shape array for shape button
BUTTON_TYPE = 5
BUTTON_ROLLOVER = 6

GRID_TYPE = 0
IMAGE_TYPE = 1

BUTTON_W = 8
BUTTON_H = 8
IMG_BUTTON_WIDTH = 300
IMG_BUTTON_HEIGHT = 50

class ConwayUI:
    def __init__(self, ui_img, ui_rollover_img, ui_selected_img, pixel_size):
        self.ui_img = ui_img
        self.ui_rollover_img = ui_rollover_img
        self.active_img = ui_img
        self.ui_selected_img = ui_selected_img
        self.pixel_size = pixel_size
        self.buttons = []
        self.selected_group_id = 0
        self.selected_button_id = -1

    def set_selected_button(self, button_id):
        self.selected_button_id = button_id
        
    def create_grid_button(self, ui_x, ui_y, shape, group_id):
        new_button_id = len(self.buttons)
        self.buttons.append([(new_button_id, group_id, ui_x, ui_y, shape, GRID_TYPE)])
        if self.selected_button_id == -1:
            self.selected_button_id = new_button_id
        return new_button_id   

    def create_image_button(self, ui_x, ui_y, img, rollover, group_id):
        new_button_id = len(self.buttons)
        self.buttons.append([(new_button_id, group_id, ui_x, ui_y, img, IMAGE_TYPE, rollover)])
        return new_button_id   

    def set_group_visibility(self, group_id):
        self.selected_group_id = group_id
        
    def draw_grid_button(self, button, game_img, mouse_x, mouse_y):        
        shape = button[BUTTON_SHAPE]

        # Show regular button or rollover?
        if button[BUTTON_ID] == self.selected_button_id:
            self.active_img = self.ui_selected_img
        elif self.check_grid_button_mouseover(button, mouse_x, mouse_y):
            self.active_img = self.ui_rollover_img
        else:
            self.active_img = self.ui_img

        # Draw the box
        for x in range(0, BUTTON_W):
            game_img.blit(self.active_img, (button[BUTTON_X] + (x * self.pixel_size), button[BUTTON_Y]))
        for x in range(0, BUTTON_W):
            game_img.blit(self.active_img, (button[BUTTON_X] + (x * self.pixel_size), button[BUTTON_Y] + (BUTTON_H * self.pixel_size)))        
        for y in range(0, BUTTON_H):
            game_img.blit(self.active_img, (button[BUTTON_X], button[BUTTON_Y] + (y * self.pixel_size)))
        for y in range(0, BUTTON_H + 1):
            game_img.blit(self.active_img, (button[BUTTON_X] + ((BUTTON_W) * self.pixel_size), button[BUTTON_Y] + (y * self.pixel_size)))

        # Draw the shape
        for y in range(0, 5):
            for x in range(0, 5):
                if shape[x][y] > 0:
                    game_img.blit(self.active_img, (button[BUTTON_X] + ((x + 2) * self.pixel_size), button[BUTTON_Y] + ((y + 2) * self.pixel_size)))

    def draw_image_button(self, button, game_img, mouse_x, mouse_y):
        if self.check_image_button_mouseover(button, mouse_x, mouse_y):
            game_img.blit(button[BUTTON_ROLLOVER], (button[BUTTON_X], button[BUTTON_Y]))
        else:
            game_img.blit(button[BUTTON_IMAGE], (button[BUTTON_X], button[BUTTON_Y]))

    def check_grid_button_mouseover(self, button, mouse_x, mouse_y):
        if (mouse_x >= button[BUTTON_X] and mouse_x < button[BUTTON_X] + (self.pixel_size * 8) + 8
            and mouse_y > button[BUTTON_Y] and mouse_y < button[BUTTON_Y] + (self.pixel_size * 8) + 8):
                return True
        else:
            return False

    def check_image_button_mouseover(self, button, mouse_x, mouse_y):
        if (mouse_x >= button[BUTTON_X] and mouse_x < button[BUTTON_X] + IMG_BUTTON_WIDTH
            and mouse_y > button[BUTTON_Y] and mouse_y < button[BUTTON_Y] + IMG_BUTTON_HEIGHT):
                return True
        else:
            return False
    
    def check_buttons(self, mouse_x, mouse_y, mouse_click):                   
        for button in self.buttons:
            if button[0][BUTTON_GROUP] == self.selected_group_id:
                if button[0][BUTTON_TYPE] == GRID_TYPE:
                    if self.check_grid_button_mouseover(button[0], mouse_x, mouse_y) and mouse_click[0] == True:                    
                        self.selected_button_id = button[0][BUTTON_ID]
                        return button[0][BUTTON_ID]
                else:
                    if self.check_image_button_mouseover(button[0], mouse_x, mouse_y) and mouse_click[0] == True:                    
                        return button[0][BUTTON_ID]        
        return -1

    def draw_all(self, game_img, mouse_x, mouse_y):
        for button in self.buttons:
            if button[0][BUTTON_GROUP] == self.selected_group_id:
                if button[0][BUTTON_TYPE] == GRID_TYPE:
                    self.draw_grid_button(button[0], game_img, mouse_x, mouse_y)
                else:
                    self.draw_image_button(button[0], game_img, mouse_x, mouse_y)
